"""
language model logic
ziyang.jia@rutgers.edu
"""
import os
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
import copy, time
from pathlib import Path

from models.data import VerseDataset
from models.model import *
from util import Util

def makeModel(config):
    model = Glyph(
        Encoder(config), # Multi-layer transformer encoder
        Generator(config)) # Predict verb
    return model

def makeOptimizer(model, config): # <<< select proper optimizer
    return NoamOpt(config, torch.optim.Adam(model.parameters(), \
            lr=0, betas=(0.9, 0.98), eps=1e-9))

def main():
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.enabled = True
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    config = Config()
    model = makeModel(config).to(device)
    criterion = LabelSmoothing(config).to(device)

    # parameter set name
    para_set_name = "_".join([str(config.batch_size), \
                str(config.batch_accumulate_size), \
                str(config.hidden_size), \
                str(config.encoder_layer_num), \
                str(config.ff_dim), \
                str(config.dropout), \
                str(config.label_smoothing), \
                str(config.optimizer_factor)])

    # log file path
    log_path = config.log_folder + para_set_name + "/"
    train_loss_path = log_path + "train_loss.out"
    train_sample_path = log_path + "train_sample.out"
    eval_loss_path = log_path + "eval_loss.out"
    eval_sample_path = log_path + "eval_sample.out"
    if not os.path.exists(log_path):
        os.mkdir(log_path)

    # load previous state_dict if exists
    state_dict_path_prefix = config.state_dict_folder + para_set_name
    if Path(state_dict_path_prefix+"_"+str(config.epoch_done-1)).is_file():
        print("LOADING STATE DICT FILE...")
        model.load_state_dict(torch.load(state_dict_path_prefix+"_"+str(config.epoch_done-1)))
        print("DONE.")
    else:
        raise Exception("ERROR: STATE DICT NOT FOUND.")

    # load data
    print("LOADING DATA...")
    eval_dataset = VerseDataset(config, data_type="eval")
    data_origin = np.load(config.data_origin_path, allow_pickle=True)
    print("DONE.")

    vocabs_dict = np.load(config.vocab_dict_path, allow_pickle=True).tolist()
    vocabs = list(vocabs_dict.keys())
    vocab_embed_dict = np.load(config.vocab_embed_dict_path, allow_pickle=True).tolist()
    vocab_glyph_embed_dict = np.load(config.vocab_glyph_embed_dict_path, allow_pickle=True).tolist()

    glyph_similarity = GlyphSimilarity(config)

    model.eval()
    eval_dataloader = DataLoader(eval_dataset, \
        batch_size=config.batch_size, shuffle=False, num_workers=0)

    log_softmax = nn.LogSoftmax(dim=-1)

    glyph_candidates = []

    # clear evaluation output
    if os.path.exists(config.glyph_out_path):
        os.remove(config.glyph_out_path)

    for i, [src, tgt, tenor, vehicle, idx] in enumerate(eval_dataloader):
        if i == len(eval_dataloader)-1:
            break

        # input & output
        src, tgt = src.to(device), tgt.to(device)
        tenor, vehicle = tenor.to(device), vehicle.to(device)
        idx = idx.to(device)
        out = model.forward(idx, src, tenor, vehicle)

        # enrich with glyph
        lang_scores, lang_idxs = out.topk(dim=-1, k=config.lang_candidate_num)
        lang_scores = lang_scores.softmax(dim=-1)
        glyph_scores = glyph_similarity.getCosineSimilarity(vehicle, lang_idxs)

        glyph_scores = glyph_scores.to(device)
        glyph_scores = glyph_scores.softmax(dim=-1)

        # try sum the scores and mul
        scores_sum = 0.001*lang_scores + glyph_scores
        scores_mul = lang_scores.mul(lang_scores).mul(glyph_scores).mul(glyph_scores)

        _, lang_top_idxs = lang_scores.topk(dim=-1, k=config.glyph_candidate_num)
        _, glyph_top_idxs = glyph_scores.topk(dim=-1, k=config.glyph_candidate_num)
        _, sum_top_idxs = scores_sum.topk(dim=-1, k=config.glyph_candidate_num)
        _, mul_top_idxs = scores_mul.topk(dim=-1, k=config.glyph_candidate_num)

        # save the output
        with open(config.glyph_out_path, "a") as glyph_out_f:
            for j in range(config.batch_size):
                scores_lang_candidates = [vocabs[lang_idxs[j][k]] for k in lang_top_idxs[j]]
                scores_glyph_candidates = [vocabs[lang_idxs[j][k]] for k in glyph_top_idxs[j]]
                scores_sum_candidates = [vocabs[lang_idxs[j][k]] for k in sum_top_idxs[j]]
                scores_mul_candidates = [vocabs[lang_idxs[j][k]] for k in mul_top_idxs[j]]

                glyph_out_f.write(str(data_origin[idx[j]])+"\n")
                glyph_out_f.write(str(scores_lang_candidates)+"\n")
                glyph_out_f.write(str(scores_glyph_candidates)+"\n")
                glyph_out_f.write(str(scores_sum_candidates)+"\n")
                glyph_out_f.write(str(scores_mul_candidates)+"\n\n")

                glyph_candidates.append(scores_sum_candidates)

    np.save("./data/snapshot/glyph_candidates.npy", glyph_candidates)

if __name__ == "__main__":
    main()
