"""
language model logic
ziyang.jia@rutgers.edu
"""
import os
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
import copy, time
from pathlib import Path

from models.data import VerseDataset
from models.model import *
from util import Util

def makeModel(config):
    model = Glyph(
        Encoder(config), # Multi-layer transformer encoder
        Generator(config)) # Predict verb
    return model

def makeOptimizer(model, config): # <<< select proper optimizer
    return NoamOpt(config, torch.optim.Adam(model.parameters(), \
            lr=0, betas=(0.9, 0.98), eps=1e-9))

def main():
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.enabled = True
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    config = Config()
    model = makeModel(config).to(device)
    criterion = LabelSmoothing(config).to(device)

    # parameter set name
    para_set_name = "_".join([str(config.batch_size), \
                str(config.batch_accumulate_size), \
                str(config.hidden_size), \
                str(config.encoder_layer_num), \
                str(config.ff_dim), \
                str(config.dropout), \
                str(config.label_smoothing), \
                str(config.optimizer_factor)])

    # log file path
    log_path = config.log_folder + para_set_name + "/"
    train_loss_path = log_path + "train_loss.out"
    train_sample_path = log_path + "train_sample.out"
    eval_loss_path = log_path + "eval_loss.out"
    eval_sample_path = log_path + "eval_sample.out"
    if not os.path.exists(log_path):
        os.mkdir(log_path)

    # load previous state_dict if exists
    state_dict_path_prefix = config.state_dict_folder + para_set_name
    if Path(state_dict_path_prefix+"_"+str(config.epoch_done-1)).is_file():
        print("LOADING STATE DICT FILE...")
        model.load_state_dict(torch.load(state_dict_path_prefix+"_"+str(config.epoch_done-1)))
        print("DONE.")

    # load optimizer with warmup
    optimizer = makeOptimizer(model, config)
    optimizer.zero_grad()

    # load data
    print("LOADING DATA...")
    train_dataset = VerseDataset(config, data_type="train")
    eval_dataset = VerseDataset(config, data_type="eval")
    data_origin = np.load(config.data_origin_path, allow_pickle=True)
    lookup_candidates = LookupCandidates(config)
    print("DONE.")

    for epoch in range(config.epoch_done, config.epoch_total):
        """
        TRAINING
        """
        print("TRAINING")
        model.train()
        train_dataloader = DataLoader(train_dataset, \
            config.batch_size, shuffle=True, num_workers=0)

        start_time = time.time()
        loss_log = []

        for i, [src, tgt, tenor, vehicle, idx] in enumerate(train_dataloader):
            if i == len(train_dataloader)-1: # drop remaining(<=batch size)
                break
            # input & output
            src, tgt = src.to(device), tgt.to(device)
            tenor, vehicle = tenor.to(device), vehicle.to(device)
            idx = idx.to(device)
            out = model.forward(idx, src, tenor, vehicle)

            # accumulate gradient
            if i%config.batch_accumulate_size == 0:
                optimizer.step()
                optimizer.zero_grad()
                out_probs, out_idxs, out_characters = lookup_candidates.getTopKCandidates(out)
                # sample output of training
                with open(train_sample_path, "a") as train_sample_f:
                    train_sample_f.write("Epoch {:d}, index {:d}:\n".format(
                        epoch, i//config.batch_accumulate_size))
                    for j, sample_idx in enumerate(idx):
                        train_sample_f.write(str(data_origin[sample_idx])+"\n")
                        train_sample_f.write(str(out_characters[j])+"\n")
                        train_sample_f.write(str(out_probs[j].exp())+"\n\n")

            # loss
            loss = criterion(out, tgt)
            loss.sum().backward()
            loss_log.append(int(loss.data.sum()))

            # print process
            if i%10 == 0:
                time_consumed = time.time()-start_time
                Util.printProgressBar(i+5, len(train_dataloader), prefix="Epoch {:d}:".format(epoch), \
                time_consumed=str(time_consumed)[:5]+"s", \
                    length=50, complete_message="EPOCH {:d} DONE.".format(epoch))
        print()

        # record the loss log & save state dict after each epoch
        with open(train_loss_path, "a") as loss_log_f:
            loss_log_f.write("Epoch {:d}:\n".format(epoch))
            for l in loss_log:
                loss_log_f.write("{:d}\n".format(l))
        torch.save(model.state_dict(), state_dict_path_prefix+"_"+str(epoch))

        """
        EVALUATION
        """
        print("EVALUATION")
        model.eval()
        eval_dataloader = DataLoader(eval_dataset, \
            batch_size=config.batch_size, shuffle=True, num_workers=0)

        start_time = time.time()
        loss_log = []

        for i, [src, tgt, tenor, vehicle, idx] in enumerate(eval_dataloader):
            if i == len(eval_dataloader)-1: # drop remaining(<=batch size)
                break

            # input & output
            src, tgt = src.to(device), tgt.to(device)
            tenor, vehicle = tenor.to(device), vehicle.to(device)
            idx = idx.to(device)
            out = model.forward(idx, src, tenor, vehicle)

            # loss
            loss = criterion(out, tgt)
            loss_log.append(int(loss.data.sum()))

            # sample output of evaluation
            out_probs, out_idxs, out_characters = lookup_candidates.getTopKCandidates(out)
            with open(eval_sample_path, "a") as eval_sample_f:
                eval_sample_f.write("Epoch {:d}, index {:d}:\n".format(
                    epoch, i))
                for j, sample_idx in enumerate(idx):
                    eval_sample_f.write(str(data_origin[sample_idx])+"\n")
                    eval_sample_f.write(str(out_characters[j])+"\n")
                    eval_sample_f.write(str(out_probs[j].exp())+"\n\n")

        print("Mean of loss: {:f}".format(sum(loss_log)/len(loss_log)))

        # record the loss log of eval
        with open(eval_loss_path, "a") as loss_log_f:
            loss_log_f.write("Epoch {:d}:\n".format(epoch))
            for l in loss_log:
                loss_log_f.write("{:d}\n".format(l))

if __name__ == "__main__":
    main()
