"""
encoder - generator
ziyang.jia@rutgers.edu
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch import matmul as mm
import numpy as np
import math

class Config(object):
    """Configurations"""
    def __init__(self, is_generator=False):
        super(Config, self).__init__()
        self.epoch_total = 100
        self.epoch_done = 72 # <<< opt for metaphor generator
        self.batch_size = 50
        self.batch_accumulate_size = 50
        # accumulated batch size: self.batch_size*self.batch_accumulate_size
        self.max_position_size = 25
        self.data_train_len = 43012

        self.embed_len = 300
        self.hidden_size = 128
        self.vocab_size = 197103

        self.encoder_layer_num = 6

        self.optimizer_warmup = 5000
        self.optimizer_factor = 2

        self.head_num = 8
        self.head_dim = 25
        self.head_sqrtd = math.sqrt(self.head_dim)

        self.ff_dim = 1024
        self.dropout = 0.1
        self.label_smoothing = 0.1

        self.mask_idx = 187974
        self.pad_idx = 187973

        self.sample_candidate_num = 10
        self.lang_candidate_num = 300 # candidates filtered by language model
        self.glyph_candidate_num = 300 # candidates filtered by glyph model

        self.snapshot_path = "./data/snapshot/"
        self.data_train_path = self.snapshot_path+"data_train.npy"
        self.data_eval_path = self.snapshot_path+"data_eval.npy"
        self.data_origin_path = self.snapshot_path+"data_origin.npy"
        self.vocab_dict_path = self.snapshot_path+"vocab_dict.npy"
        self.vocab_embed_dict_path = self.snapshot_path+"vocab_embed_dict.npy"
        self.vocab_glyph_embed_dict_path = self.snapshot_path+"vocab_glyph_embed_dict.npy"

        self.state_dict_folder = "./data/state_dict/"
        self.log_folder = "./log/"

        self.glyph_out_path = "./data/output/glyph.out"

        if is_generator is True:
            self.epoch_done = self.epoch_total

class Glyph(nn.Module):
    """ Multilayer Encoder + Emphasizing Metephor """
    def __init__(self, encoder, generator):
        super(Glyph, self).__init__() # parameters are initialized in the model
        self.encoder = encoder
        self.generator = generator

    def forward(self, idx, src, tenor, vehicle):
        return self.generator(idx, src, self.encoder(src, tenor, vehicle))
        
class Encoder(nn.Module):
    """Transformer multi-layer Encoder structure"""
    def __init__(self, config):
        super(Encoder, self).__init__()
        self.embeddings = Embeddings(config)
        self.embed_feedforward = FeedForward(config, is_embed=True)
        self.layers = nn.ModuleList([
            EncoderLayer(config) for _ in range(config.encoder_layer_num)])
        self.norm = LayerNorm(config.hidden_size)

        for module in [self.embed_feedforward, self.layers, self.norm]:
            initParameters(module)

    def forward(self, src, tenor, vehicle):
        h = self.embed_feedforward(self.embeddings(src, tenor, vehicle))
        for layer in self.layers:
            h = layer(h)

        return self.norm(h)

class EncoderLayer(nn.Module):
    """Layer sturcture according to Transformer encoder and Havard annotation"""

    def __init__(self, config):
        super(EncoderLayer, self).__init__()
        self.self_attention = SelfAttention(config)
        self.feed_forward = FeedForward(config)
        self.norm = LayerNorm(config.hidden_size)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, x):
        # SelfAttention sublayer
        y = x + self.dropout(self.self_attention(self.norm(x)))
        # Feedforward sublayer
        y = y + self.dropout(self.feed_forward(self.norm(y)))

        return y

class SelfAttention(nn.Module):
    """Multiheaded SelfAttention"""
    def __init__(self, config):
        super(SelfAttention, self).__init__()
        self.d_qkv = config.head_dim
        self.w_qkvs = nn.ModuleList([   # config.head_num * [qkv]
            nn.ModuleList([
                nn.Linear(config.hidden_size, self.d_qkv) 
                for _ in range(3)]) 
            for _ in range(config.head_num)])
        self.w_ff = nn.Linear(config.head_num*config.head_dim, config.hidden_size)
        self.softmax = nn.Softmax(dim=-1)
        self.head_num = config.head_num
        self.head_sqrtd = config.head_sqrtd

    def forward(self, x):
        v_sums = [0]*self.head_num
        for i in range(self.head_num):
            q, k, v = [l(x) for l in self.w_qkvs[i]]
            v_sums[i] += self.softmax(mm(q.permute(0, 2, 1),k)/self.head_sqrtd)*v
        return self.w_ff(torch.cat(v_sums, -1)) # to save time, just concate once in return

class FeedForward(nn.Module):
    """docstring for FeedForward"""
    def __init__(self, config, is_embed=False):
        super(FeedForward, self).__init__()
        if is_embed:
            self.l1 = nn.Linear(config.embed_len, config.ff_dim)
            self.l2 = nn.Linear(config.ff_dim, config.hidden_size)
            self.dropout = nn.Dropout(config.dropout)
        else:
            self.l1 = nn.Linear(config.hidden_size, config.ff_dim)
            self.l2 = nn.Linear(config.ff_dim, config.hidden_size)
            self.dropout = nn.Dropout(config.dropout)

    def forward(self, x):
        return self.l2(self.dropout(F.relu(self.l1(x))))

class LayerNorm(nn.Module): # <<< can be replaced by pytorch nn.LayerNorm
    """docstring for LayerNorm"""
    def __init__(self, features_dim, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.a = nn.Parameter(torch.ones(features_dim))
        self.b = nn.Parameter(torch.zeros(features_dim))
        self.eps = eps
    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a * (x-mean) / (std+self.eps) + self.b

class Embeddings(nn.Module):
    """Read pre-trained embeddings from glove file (word-vector)"""
    def __init__(self, config):
        super(Embeddings, self).__init__()
        vocab_embed_dict = np.load(config.vocab_embed_dict_path, allow_pickle=True).tolist()
        vocab_dict = np.load(config.vocab_dict_path, allow_pickle=True).tolist()
        assert list(vocab_embed_dict.keys()) == list(vocab_dict.keys())

        weight_mat = np.zeros((len(vocab_embed_dict.keys()), config.embed_len))
        for i, w in enumerate(list(vocab_embed_dict.keys())):
            weight_mat[i] = vocab_embed_dict[w]
        self.word_embed = nn.Embedding(*weight_mat.shape)
        self.word_embed.load_state_dict({"weight": torch.from_numpy(weight_mat)})
        self.word_embed.weight.requires_grad = False

        self.post_embed = PositionEmbed(config.max_position_size, config.embed_len)

        self.norm = LayerNorm(config.embed_len)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, src, tenor, vehicle): # including tenor and vehicle in experiments
        src_pos = torch.arange(src.size(-1), dtype=torch.long, device=src.device)
        if len(src) == 1:
            src_pos = src_pos.expand_as(src)
        else:
            src_pos = src_pos.unsqueeze(0).expand_as(src)

        src_embed = self.word_embed(src)
        src_pos_embed = self.post_embed(src_pos)

        embeddings = src_embed+src_pos_embed
        embeddings = self.norm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings

class PositionEmbed(nn.Module):
    """Position embedding according to attention is all you need"""
    def __init__(self, x_dim, y_dim):
        super(PositionEmbed, self).__init__()
        post_embed = torch.zeros(x_dim, y_dim)
        position = torch.arange(0., x_dim).unsqueeze(1)
        t = torch.exp(torch.arange(0., y_dim, 2) * -(math.log(10000.0) / y_dim))
        post_embed[:, 0::2] = torch.sin(position*t)
        post_embed[:, 1::2] = torch.cos(position*t)
        post_embed = post_embed.unsqueeze(0)
        self.register_buffer("post_embed", post_embed)

    def forward(self, x):
        return Variable(self.post_embed[:, :x.size(-1)],
                            requires_grad=False)

class NoamOpt:
    "Optim wrapper that implements rate."
    def __init__(self, config, optimizer):
        self.optimizer = optimizer
        self._step = config.epoch_done*(config.data_train_len//(config.batch_size*config.batch_accumulate_size))
        self.warmup = config.optimizer_warmup
        self.factor = config.optimizer_factor
        self.hidden_size = config.hidden_size
        self._rate = 0
        
    def step(self):
        "Update parameters and rate"
        self._step += 1
        rate = self.rate()
        for p in self.optimizer.param_groups:
            p['lr'] = rate
        self._rate = rate
        # self._rate = 0.01   # <<<
        self.optimizer.step()

    def rate(self, step = None):
        "Implement `lrate` above"
        if step is None:
            step = self._step
        return self.factor * \
            (self.hidden_size ** (-0.5) *
            min(step ** (-0.5), step * self.warmup ** (-1.5)))

    def zero_grad(self):
        "Clear gradients"
        self.optimizer.zero_grad()

class LabelSmoothing(nn.Module):
    def __init__(self, config):
        super(LabelSmoothing, self).__init__()
        self.smoothing = config.label_smoothing
        self.class_num = config.vocab_size
        self.criterion = nn.KLDivLoss(reduction="none")
        self.confidence = 1.0 - config.label_smoothing
        self.confidence_not = self.smoothing / (self.class_num - 2)  # 2 for tgt and pad
        self.pad_idx = config.pad_idx
        
    def forward(self, out, tgt):
        true_dist = torch.zeros_like(out)
        true_dist.fill_(self.confidence_not)
        true_dist.scatter_(1, tgt.data.unsqueeze(1), self.confidence)
        true_dist[:, self.pad_idx] = 0

        return self.criterion(out, Variable(true_dist, requires_grad=False))

class GlyphSimilarity(object):
    """docstring for GlyphSimilarity"""
    def __init__(self, config):
        super(GlyphSimilarity, self).__init__()
        self.cs = nn.CosineSimilarity(dim=-1)
        self.batch_size = config.batch_size
        self.vocab_glyph_embed_dict = np.load(config.vocab_glyph_embed_dict_path, allow_pickle=True).tolist()
        self.vocabs = list(self.vocab_glyph_embed_dict.keys())

    def getCosineSimilarity(self, vehicle_batch, candidates_batch):
        glyph_similaritys = []
        for i in range(self.batch_size):
            vehicle = self.vocabs[vehicle_batch[i]]
            vehicle_glyph_embed = self.vocab_glyph_embed_dict.get(vehicle)
            glyph_similarity = []
            candidates = [self.vocabs[c_idx] for c_idx in candidates_batch[i]]

            candidates_cosine_similarity = torch.tensor([ \
                    self.cs(self.vocab_glyph_embed_dict.get(c), vehicle_glyph_embed) \
                        for c in candidates
                ])
            glyph_similaritys.append(candidates_cosine_similarity)
        return torch.stack(glyph_similaritys)

class Generator(nn.Module):
    """ linear + softmax """
    def __init__(self, config):
        super(Generator, self).__init__()
        self.mask_idx = config.mask_idx
        self.hidden_size = config.hidden_size
        self.l = nn.Linear(config.hidden_size, config.vocab_size)

        initParameters(self.l)
        
    def forward(self, idx, src, x):
        mask = torch.nonzero(src == self.mask_idx)[:, 1]
        assert mask.shape[0] == src.shape[0], \
            "mask shape: {:s}, src shape: {:s}".format(str(mask.shape), str(src.shape))
        t_x = torch.gather(x, 1, mask.view(-1, 1).repeat(1,1,self.hidden_size).reshape(-1, 1, self.hidden_size))
        return F.log_softmax(self.l(torch.squeeze(t_x, 1)), dim=-1)

def initParameters(module):
    for p in module.parameters():
        if p.dim()>1:
            nn.init.xavier_uniform_(p)

class LookupCandidates(object):
    """lookup top n candidates through log_softmax"""
    def __init__(self, config):
        super(LookupCandidates, self).__init__()
        self.vocabs = list(np.load(config.vocab_dict_path, allow_pickle=True).tolist().keys())
        self.n = config.sample_candidate_num
        self.mask_idx = config.mask_idx

    def getTopKCandidates(self, x):
        probs, idxs = x.topk(self.n, largest=True, sorted=True)
        characters = []
        for candidate_idxs in idxs:
            characters.append([self.vocabs[idx] for idx in candidate_idxs])
        return probs, idxs, characters