"""
language model logic
ziyang.jia@rutgers.edu
"""
import os
import re
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
import copy, time
from pathlib import Path
from pypinyin import pinyin, Style

from models.data import VerseDataset
from models.model import *
from util import Util

# According to 新编诗韵
class CharacterRhymeIdx(object):
    """docstring for CharacterRhymeIdx"""
    def __init__(self):
        super(CharacterRhymeIdx, self).__init__()
        self.character_rhyme_dict = np.load("./data/snapshot/character_rhyme_dict.npy", allow_pickle=True).tolist()
        self.rhyme_character = np.load("./data/snapshot/rhyme_character.npy", allow_pickle=True).tolist()

    def lookupRhyme(self, c):
        if self.character_rhyme_dict.get(c) is None:
            with open("./temp_out", "a") as t_f:
                t_f.write(c)
            return 0
        return self.character_rhyme_dict[c]

class RhymePattern(object):
    """docstring for RhymePattern"""
    def __init__(self, character_rhyme_idx):
        super(RhymePattern, self).__init__()
        self.character_rhyme_idx = character_rhyme_idx

    def getPattern(self, endings, e_idx):
        if e_idx == 0:
            # [A]BA
            if len(endings) >= 3:
                r_idx = self.character_rhyme_idx.lookupRhyme(endings[2])
                r_span = [0, len(endings)-1]
                r_des = "[A]BA"
            # [A]A
            elif len(endings) == 2:
                r_idx = self.character_rhyme_idx.lookupRhyme(endings[1])
                r_span = [0, len(endings)-1]
                r_des = "[A]A"
            else:
                raise Exception("ERROR: one line cannot build a rhyme.")
        elif e_idx == 1:
            # A[B]CB
            if len(endings) >=4:
                r_idx = self.character_rhyme_idx.lookupRhyme(endings[3])
                r_span = [0, 3]
                r_des = "A[B]CB"
            # A[A]
            else:
                r_idx = self.character_rhyme_idx.lookupRhyme(endings[0])
                r_span = [0, len(endings)-1]
                r_des = "A[A]"
        elif e_idx == len(endings)-1:
            # AB[A]
            if len(endings) >= 3:
                r_idx = self.character_rhyme_idx.lookupRhyme(endings[-1])
                r_span = [len(endings)-3, len(endings)-1]
                r_des = "AB[A]"
            else:
                raise Exception("ERROR: e_idx == len(endings)-1")
        elif e_idx == len(endings)-2:
            # AB[A]C
            if len(endings) >= 4:
                r_idx = self.character_rhyme_idx.lookupRhyme(endings[-1])
                r_span = [len(endings)-4, len(endings)-1]
                r_des = "AB[A]C"
            else:
                raise Exception("e_idx == len(endings)-2")
        else:
            r_idx_0 = self.character_rhyme_idx.lookupRhyme(endings[e_idx-2])
            r_idx_1 = self.character_rhyme_idx.lookupRhyme(endings[e_idx-1])
            r_idx_3 = self.character_rhyme_idx.lookupRhyme(endings[e_idx+1])
            r_idx_4 = self.character_rhyme_idx.lookupRhyme(endings[e_idx+2])

            r_span = [e_idx-2, e_idx+2]

            # AB[A]CA
            if r_idx_0 == r_idx_4:
                r_idx = r_idx_0
                r_des = "AB[A]BA"
            # AB[B]BC
            elif r_idx_1 == r_idx_3:
                r_idx = r_idx_1
                r_des = "AB[B]BC"
            # AB[A]CD
            else:
                r_idx = r_idx_0
                r_des = "AB[A]CD"

        return [r_idx, r_span, r_des]
        

def main():
    # preprocessing
    os.system("python ./data/rhyme/format_data.py")

    config = Config()
    character_rhyme_idx = CharacterRhymeIdx()
    rhyme_pattern = RhymePattern(character_rhyme_idx)

    # load data
    data_origin = np.load(config.data_origin_path, allow_pickle=True)
    data_eval = np.load(config.data_eval_path, allow_pickle=True)
    glyph_candidates = np.load("./data/snapshot/glyph_candidates.npy", allow_pickle=True)

    if os.path.exists("./data/snapshot/eval_rhyme_idxs.npy"):
        eval_rhyme_idxs = np.load("./data/snapshot/eval_rhyme_idxs.npy")
    else:
        eval_rhyme_idxs = []
        for i, data in enumerate(data_eval):
            do = data_origin[data["id"]]
            src_str = ""
            for w in do["src"]:
                if w == "[MASK]":
                    src_str += w
                elif w == "[PAD]":
                    break
                else:
                    ccs = re.findall(r'[\u4e00-\u9fff]+', w)
                    if len(ccs) == 0:
                        continue
                    else:
                        src_str += w
            if src_str[-6:] == "[MASK]":
                eval_rhyme_idxs.append(i)
        np.save("./data/snapshot/eval_rhyme_idxs.npy", np.array(eval_rhyme_idxs))

    rhyme_predictions = []

    for i, eval_idx in enumerate(eval_rhyme_idxs):
        data = data_origin[data_eval[eval_idx]["id"]]
        endings = data["endings"]
        e_idx = data["line_idx"]
        [r_idx, r_span, r_des] = rhyme_pattern.getPattern(endings, e_idx)

        for w in glyph_candidates[eval_idx]:
            w_r_idx = character_rhyme_idx.lookupRhyme(w[-1])
            if w_r_idx == r_idx:
                print(i)
                rhyme_predictions.append({
                    "w": w,
                    "span": r_span,
                    "des": r_des,
                })
                break
        if (i+1) < len(rhyme_predictions):
            print("Warning: Rhyme word not found.")
            rhyme_predictions.append({
                "w": glyph_candidates[eval_idx][0],
                "span": [e_idx, e_idx+1],
                "des": "Rhyme word not found.",
            })

    np.save("./data/snapshot/rhyme_predictions.npy", rhyme_predictions)

if __name__ == "__main__":
    main()
